/*******************************************************************************
*  Berzerk.cpp - Main entry point for Berzerk
*
*  Berzerk/
*
*  Part of the BuckoSoft Network Monitor
*  Copyright(c) 2008 - BuckoSoft, Inc.
*
*  http://git.buckosoft.com/gitweb/?p=Berzerk.git;a=history;f=Berzerk/Berzerk.cpp
*
*  With which we monitor many log files and perform appropriate action.
*
*  Debug levels:
*  0 = off
*  1,2,3,4	config stuff
*  5,6		matched lines
*  7,8,9	full output
\*******************************************************************************/

#include "stdafx.h"
#include "../config.h"

#if 0
#include <xercesc/util/PlatformUtils.hpp>
#endif

#include <boost/asio.hpp>
#include <boost/thread.hpp>
#include <boost/bind.hpp>
#include <boost/regex.hpp>
#include <boost/foreach.hpp>
#include <boost/program_options.hpp>

#include <iostream>
#include <fstream>
#include <string>
#include <syslog.h>
#include <list>

#include <pthread.h>
#include <signal.h>
#include "server/server.hpp"


//#include <iterator>
//#include <algorithm>
#include "Filter.h"
#include "Berzerk.h"
#include "WebHandler.h"

#ifdef	UNIX
#include <syslog.h>
#endif

using namespace boost;
namespace po = boost::program_options;

int	_logLevel = 1;
int	_serverPort = -1;

PCSTR	primaryConfigFileName = "berzerk.bzf";

namespace berzerk {
Berzerk	berzerk;
int	DEBUG = 0;
}
#if 0
using namespace xercesc;
#endif

// forward decls

///////////////////////////////////////////////////////////////////////////////
int _tmain(int argc, _TCHAR* argv[]) {

	int ret = berzerk::berzerk.parseCommandLine(argc, argv);
	if (ret != 0)
		exit(ret);
	berzerk::berzerk.prepareRegexp();
	berzerk::berzerk.openLogFiles();
	if (berzerk::berzerk.isDaemonMode()) {
		ret = daemon(1,0);
		berzerk::berzerk.writePidFile();
	}
#if 0
	try {
		XMLPlatformUtils::Initialize();
	}
	catch (const XMLException& toCatch) {
		// Do your failure processing here
		return 1;
	}
#endif

    // Block all signals for background thread.
    sigset_t new_mask;
    sigfillset(&new_mask);
    sigset_t old_mask;
    pthread_sigmask(SIG_BLOCK, &new_mask, &old_mask);

    berzerk::WebHandler	webHandler;
    webHandler.setStatistics(&berzerk::berzerk.getStatistics());

    // Run server in background thread.
    if (berzerk::DEBUG > 0)
    	cout << "Opening server listen on " << berzerk::berzerk.getServerAddress() << ":" << berzerk::berzerk.getServerPort() << endl;
    ::http::server::server s(berzerk::berzerk.getServerAddress(), berzerk::berzerk.getServerPort(), "/home/dick/tmp", webHandler);
    ::boost::thread t(boost::bind(&::http::server::server::run, &s));

    // Restore previous signals.
    pthread_sigmask(SIG_SETMASK, &old_mask, 0);


    berzerk::berzerk.run();

    // Stop the server.
    s.stop();
    t.join();

#if 0
    XMLPlatformUtils::Terminate();
#endif
    return(0);
}

namespace berzerk {
///////////////////////////////////////////////////////////////////////////////
Berzerk::Berzerk() {
	ipTables = false;
	bSyslog = false;
	dryRun = false;
	daemonMode = false;
	tailMode = false;
	config.setBerzerk(this);
	serverAddress = "127.0.0.1";
	serverAddress = "0.0.0.0";
	//serverPort = "9090";
}
Berzerk::~Berzerk() {
	for (map<string, BadGuy*>::iterator iter = badguyMap.begin(); iter != badguyMap.end(); iter++) {
		delete iter->second;
	}
}

///////////////////////////////////////////////////////////////////////////////
void Berzerk::usage() {
	cout << "Usage: berzerk [-n] [-t] [-iptables] [-syslog] [-f filter]" << std::endl;
}

///////////////////////////////////////////////////////////////////////////////
int Berzerk::parseCommandLine(int argc,  _TCHAR* argv[]) {
#ifdef	WIN32
	std::wstring	fileName;
#else
	std::string		fileName;
#endif

	po::options_description options("Supported options");
	options.add_options()
		("help", "produce help message")
		("iptables", po::value<bool>()->zero_tokens(), "Output to iptables")
		("syslog", po::value<bool>()->zero_tokens(), "Output to syslog")
		("daemon", po::value<bool>()->zero_tokens(), "Run as a daemon (implies 'tail -f'")
		("n", "dry run (don't alter system)")
		("t", "tail -f mode for debugging")
		("d", po::value<int>(&DEBUG)->default_value(0), "debug level 0-9")
		("l", po::value<int>(&_logLevel)->default_value(1), "log level 0-9")
		("port", po::value<int>(&_serverPort)->default_value(-1), "port that the stat server listens on")
		("config-file", po::value< vector<string> >(), "config file")
		("f", po::value<string>(), "filter file")
		;
	po::positional_options_description p;
	p.add("config-file", -1);


	po::variables_map vm;
	po::store(po::parse_command_line(argc, argv, options), vm);
	po::store(po::command_line_parser(argc, argv). options(options).positional(p).run(), vm);
	po::notify(vm);

	if (vm.count("help")) {
		usage();
	    cout << options << "\n";
	    return 1;
	}
	if (vm.count("iptables"))
		this->setIpTables(true);
	if (vm.count("syslog"))
		this->setSyslog(true);
	if (vm.count("t"))
		this->setTailMode(true);
	if (vm.count("daemon")) {
		this->setDaemonMode(true);
		this->setTailMode(true);
	}
	if (vm.count("n"))
		this->setDryRun(true);

	Filter::setDEBUG(DEBUG);
	berzerk::WebHandler::setDEBUG(DEBUG);
	Config::setDEBUG(DEBUG);
	char buf[PATH_MAX];
	log.setLogFmt(config.getLogFmt());
	log.setLogLevel(_logLevel);
	berzerk.setServerPort(_serverPort != -1 ? _serverPort : config.getPort());

	string rootDir = getcwd(buf, PATH_MAX);
	ifstream	file;
	string s = primaryConfigFileName;
	file.open(primaryConfigFileName);
	if (!file.good()) {
		file.close();
		if (DEBUG > 1)
			cout << ".. Can't open config file: " << s << endl;
		s = CONF_DIR;
		s += "/Berzerk";
		rootDir = s;
		s += "/";
		s += primaryConfigFileName;
		file.open(s.c_str());
	}
	if (!file.good()) {
		// maybe development dir...
		file.close();
		if (DEBUG > 1)
			cout << ".. Can't open config file: " << s << endl;
		s = getcwd(buf, PATH_MAX);
		s += "/etc";
		rootDir = s;
		s += "/";
		s += primaryConfigFileName;
		file.open(s.c_str());
	}
	if (file.good()) {
		if (DEBUG > 1)
			cout << ".. Opened config file: " << s << endl;
		Filter::setRootDirectory(rootDir);
		if (!config.readFile(file))
			return(2);
	} else {
		if (DEBUG > 1)
			cout << ".. Can't open config file: " << s << endl;

	}
	file.close();

	if (!vm["config-file"].empty()) {
		BOOST_FOREACH(string filename, vm["config-file"].as<vector<string> >()) {
			if (!Filter::parseFile(filename.c_str()))
				return(2);
		}
	}
	cout << "Processing: " << this->logFiles.size() << endl;
	if (this->filters.empty()) {
		PCSTR sx = "No config files specified!  Aborting.";
		cerr << sx << endl;
		cout << sx << endl;
		return(2);
	}

	return(0);
}

///////////////////////////////////////////////////////////////////////////////
void Berzerk::prepareRegexp() {
	BOOST_FOREACH(Filter* filter, filters) {
		filter->prepareRegexp();
	}
}

///////////////////////////////////////////////////////////////////////////////
void Berzerk::setServerPort(int port) {
	string s;
	char	c[32];
	sprintf(c, "%d", port);
	this->serverPort = c;
}

///////////////////////////////////////////////////////////////////////////////
void Berzerk::writePidFile() {
	pid_t	pid;
	char	buff[256];
	string	fname = VARRUN_DIR;
	fname += "/berzerk.pid";

	pid = getpid();
	snprintf(buff, 255, "%d\n", pid);
	ofstream	file;

	file.open(fname.c_str(), ios::trunc | ios::out);
	if (file.good()) {
		file << buff;
	}
	file.close();

}

///////////////////////////////////////////////////////////////////////////////
int Berzerk::openLogFiles() {
	BOOST_FOREACH(Filter* filter, filters) {
		string cfile = filter->getLogFile();
		bool found = false;
		BOOST_FOREACH(LogFile* logFile, logFiles) {
			if (logFile->name.compare(cfile) == 0) {
				found = true;
				break;
			}
		}
		if (!found) {
			LogFile* lf = new LogFile();
			if (DEBUG)
				cout << ".. openLog: '" << cfile << "'" << endl;
			lf->name = cfile;
			lf->stream.open(cfile.c_str(), ifstream::in);
			lf->pos = 0;
			if (!lf->stream.good()) {
				std::cout << "Failed to open log file '" << cfile.c_str() << "'" << std::endl;
				return(false);
			}
			logFiles.push_back(lf);
		}
	}
	return(0);
}

///////////////////////////////////////////////////////////////////////////////
int Berzerk::run() {
	std::string line;
	char		buff[4096];
	while(true) {
		bool processed = false;
		BOOST_FOREACH(LogFile* logFile, logFiles) {
			logFile->stream.seekg(0, ios::end);
			int length = logFile->stream.tellg();
			logFile->stream.seekg(logFile->pos, ios::beg);
			if (length < logFile->pos) {
				logFile->pos = 0;
				string t = "Log reset : ";
				t += logFile->name;
				log.log(2, t);
				continue;
			}
			if (length > logFile->pos) {
				if (DEBUG > 7)
					cout << "length = " << length << " pos= " << logFile->pos << endl;
				logFile->stream.getline(buff, 4095);
				logFile->pos = logFile->stream.tellg();
				if (DEBUG > 6)
					cout << "..." << buff << std::endl;
				if (logFile->stream.fail()) {
					logFile->stream.clear();
					continue;
				}
				line = buff;
				statistics.incLinesRead();
				bool old_processed = processed;
				processed = true;
				BOOST_FOREACH(Filter* filter, filters) {
					if (filter->getLogFile() == logFile->name) {
						boost::smatch matches;
						BOOST_FOREACH(regex* reg, filter->patterns) {
							if (regex_search(line, matches, *reg)) {
								boost::regex ipPattern(filter->getIPPattern());
								lineMatched(filter, line, ipPattern, reg);
							}
						}
					}
				}
				if (logFile->stream.eof())
					processed = old_processed;
			}
		}
		if (!processed) {
			if (isTailMode())
				sleep(1);
			else
				return(0);
		}
	}
	return(-1);
}

///////////////////////////////////////////////////////////////////////////////
const LogFile* Berzerk::getLogFile(int i) {
	int j=0;
	BOOST_FOREACH(LogFile* logFile, logFiles) {
		if (j == i)
			return(logFile);
		j++;
	}
	return(NULL);
}

///////////////////////////////////////////////////////////////////////////////
bool Berzerk::lineMatched(Filter* filter, const string& line, const regex& ipPattern, const regex* triggered) {
	smatch	ipmatches;
	char	buff[256];

	if (DEBUG > 4 || dryRun) {
		cout << "*************************************" << std::endl;
		cout << line << std::endl;
	}
	if (regex_search(line, ipmatches, ipPattern)) {
		if (DEBUG > 4 || dryRun)
			cout << "IP:" << "'" << ipmatches[0] << "'  Rule: '" << filter->getName() << "'" << std::endl;
	} else {
		cout << "!!!!!!!!!!!!! Can't match ip" << std::endl;
		if (!(DEBUG > 4) && !dryRun)
			cout << line << endl;
		return(false);
	}
	const string badguyIP = ipmatches[0];
	if (config.isWhiteList(badguyIP)) {
		snprintf(buff, 255, "whitelist: '%s'", badguyIP.c_str());
		if (DEBUG > 4|| dryRun)
			cout << buff << endl << endl;
		log.log(3, buff);
		log.log(3, line);
		return(false);
	}
	BadGuy* badguy;
	if (badguyMap[badguyIP] != NULL) {
		badguy = badguyMap[badguyIP];
	} else {
		badguy = new BadGuy(badguyIP, filter->getName(), 0, 0);
		badguyMap[badguyIP] = badguy;
	}
	badguy->sum += filter->getWeight();
	int	sum = badguy->sum;
	if (DEBUG > 4)
		cout << "sum for '" << badguyIP << "' is " << sum << endl;
	if (DEBUG > 4|| dryRun)
		cout << endl;
	string syslogString;
	snprintf(buff, 255, "%s: '%s' broke rule '%s' (+%d=%d)", sum < config.getThreshold() ? "Warn" : "Kill",
			 badguyIP.c_str(), filter->getName().c_str(), filter->getWeight(), sum);
	syslogString = buff;

	if (sum < config.getThreshold()) {
		log.log(2, syslogString);
		log.log(3, line);
		return(false);
	}
// XXX
	time_t when = now();
	time_t age = when - badguy->when;
	if (age < config.getTriggerDelay()) {
		if (DEBUG > 4|| dryRun)
			cout << "skip kill, age = " << age << endl;
		return(false);

	}
	badguy->when = when;

	if (filter->getTrap() == "logkill") {
		statistics.incKilled(badguyIP, filter->getName(), line);
		if (this->dryRun || DEBUG)
			cout << "::trap: logkill: " << syslogString << endl;
		if (!this->dryRun) {
			log.log(1, syslogString);
			syslog(LOG_PID | LOG_LOCAL0 | LOG_WARNING, "%s", (PCSTR)syslogString.c_str());
		}
	} else {
		// trap = [arschloch] $iptables -A arschloch -p ALL -s $ip -j DROP

	}
#ifdef	UNIX
	if (this->isSyslog()) {
		if (this->dryRun || DEBUG)
			cout << "::syslog: " << syslogString << endl;
		if (!this->dryRun) {
			log.log(1, syslogString);
			syslog(LOG_PID | LOG_LOCAL0 | LOG_WARNING, "%s", syslogString.c_str());
		}
	}
#ifdef PATH_TO_IPTABLES
	if (this->isIpTables()) {
		string	s = ipmatches[0];
		PCSTR	ip = s.c_str();
		char* args[] = {(char*)PATH_TO_IPTABLES, (char*)"-A", (char*)"arschloch", (char*)"-p", (char*)"ALL", (char*)"-s", (char*)s.c_str(), (char*)"-j", (char*)"DROP"};
		string w;
		w = PATH_TO_IPTABLES;
		w += " -A arschloch -p ALL -s ";
		w += ip;
		w += " -j DROP";
		if (this->dryRun)
			cout << "::iptables: " << w << endl;
		else {
			log.log(1, w);
			pid_t pid = fork();
			if (pid == 0)
				execv(PATH_TO_IPTABLES, args);
		}
		// `$iptables -A arschloch -p ALL -s $ip -j DROP`;
		//syslog(LOG_PID | LOG_LOCAL0 | LOG_INFO, "Kill: '%s' accessed '%s'", s.c_str(), line.c_str());

	}
#endif
	log.log(2, line);
#endif
	return(true);
}
}	// namespace berzerk
