/*******************************************************************************
*  Filter.cpp - Define a filter
*
*  Berzerk/
*
*  Part of the BuckoSoft Network Monitor
*  Copyright(c) 2008 - BuckoSoft, Inc.
*
*  http://git.buckosoft.com/gitweb/?p=Berzerk.git;a=history;f=Berzerk/Filter.cpp
*
*  $Id: Filter.cpp,v 1.8 2009/10/13 02:24:43 dick Exp $
*
\*******************************************************************************/


#include "stdafx.h"
#include "../config.h"

#include <string>
#include <iostream>
#include <fstream>

#include <boost/foreach.hpp>
#include <boost/algorithm/string.hpp>

using namespace std;
using namespace boost;

#include "Filter.h"
#include "Berzerk.h"

int Filter::DEBUG = 0;
string	Filter::rootDirectory;

///////////////////////////////////////////////////////////////////////////////
Filter::Filter(void) {
	weight = 1;
}

///////////////////////////////////////////////////////////////////////////////
Filter::~Filter(void) {
}

///////////////////////////////////////////////////////////////////////////////
// static
bool	Filter::parseFile(PCSTR filename) {
	ifstream	file;
	if (DEBUG)
		cout << "name: " << filename << "\n";
	file.open(filename, ifstream::in);
	if (!file.good()) {
		file.close();
		string s = rootDirectory;
		s += "/";
		s += filename;
		file.open(s.c_str());
	}
	if (!file.good()) {
		std::cout << "Failed to open filter file '" << filename << "'" << std::endl;
		return(false);
	}
	Filter* filter = new Filter();
	if (!filter->readFile(file))
		return(false);
	berzerk::berzerk.addFilter(filter);
	file.close();
	return(true);
}

///////////////////////////////////////////////////////////////////////////////
bool	Filter::readFile(ifstream& ifs)
{
	string		line;
	char		buff[256];

	bool	emptyLine = true;
	while(true) {
		ifs.getline(buff, 256);
		if (ifs.fail()) {
			if (this->matchPatterns.size() < 1) {
				cout << "Failed to parse filter file" << std::endl;
				return(false);
			}
			break;
		}
		line = buff;
		if (DEBUG > 5)
			cout << ">>> '" << line << "'" << endl;
		if (buff[0] == '#' || line.empty()) {
			emptyLine = true;
			continue;
		}
		if (emptyLine && buff[0] == '[') {
			line = &buff[1];
			line.erase(line.length()-1);
			if (!this->name.empty()) {
				if (DEBUG > 5)
					cout << "++ New Filter: '" << line << "'" << endl;
				Filter* filter = new Filter();
				//filter->setBerzerk(this->berzerk);
				filter->setName(line);
				if (!filter->readFile(ifs))
					return(false);
				berzerk::berzerk.addFilter(filter);
				break;
			}
			this->name = line;
			continue;
		}
		typedef vector< string > split_vector_type;
		split_vector_type sv;
		split(sv, line, is_any_of("="));
		trim(sv[0]);
		trim(sv[1]);

		if (DEBUG > 4) {
			cout << "type ='" << sv[0] << "' data = '" << sv[1] << "'" << endl;
		}

		if (sv[0] == "log") {
			if (DEBUG > 4)
				cout << "++ logFile: '" << sv[1] << "'" << endl;
			if (sv[1][0] == '@') {
				split(sv,sv[1], is_any_of("@"));
				sv[1] = berzerk::berzerk.getMappedLogFile(sv[1]);
			}
			this->logFile = sv[1];
			continue;
		}
		if (sv[0] == "trap") {
			if (DEBUG > 4)
				cout << "++ trap: '" << sv[1] << "'" << endl;
			this->trap = sv[1];
			continue;
		}
		if (sv[0] == "ip") {
			if (DEBUG > 4)
				cout << "++ ipExtract: '" << line << "'" << endl;
			this->ipExtractPattern = sv[1];
			continue;
		}
		if (sv[0] == "filter") {
			if (DEBUG > 4)
				cout << "++ matchPat: '" << line << "'" << endl;
			this->matchPatterns.push_back(sv[1]);
			continue;
		}
		if (sv[0] == "weight") {
			int i = atoi(sv[1].c_str());
			if (DEBUG > 4)
				cout << "++ weight: '" << sv[1] << "' (" << i << ")" << endl;
			this->weight = i;
			continue;
		}
		cout << "***Unknown line type '" << sv[0] << "'" << endl;
		return(false);
	}

//	cout << "pattern = '" << this->matchPatterns << "'" << endl;
	if (DEBUG > 3) {
		cout << "-----------------------------------------" << endl;
		cout << "Filter ---- '" << this->name << "'" << endl;
		cout << "  logFile = '" << this->logFile << "'" << endl;
		cout << "ipPattern = '" << this->ipExtractPattern << "'" << endl;
		BOOST_FOREACH(string pat, matchPatterns) {
			cout << "      pat = '" << pat << "'" << endl;
		}
	}
	return(true);
}

///////////////////////////////////////////////////////////////////////////////
void Filter::prepareRegexp() {
	BOOST_FOREACH(string mpat, getMatchPatterns()) {
		patterns.push_back(new regex(mpat.c_str()));
	}
}

